#!/usr/bin/env python
# -*- coding: utf-8 -*-

import pytest
from pydicom.filebase import DicomBytesIO

from dicom_utils.basic_offset_table import BYTES_PER_LONG, LENGTH, BasicOffsetTable
from dicom_utils.dicom import Dicom


@pytest.fixture
def dummy_dicom_factory(mocker):
    def func(pixels):
        m = mocker.MagicMock(Dicom)
        m.PixelData = pixels
        return m

    return func


@pytest.fixture(params=["bytes", "dicombytesio", "dicom"])
def stream_type_factory(dummy_dicom_factory, request):
    stream_type = request.param

    def func(stream):
        if stream_type == "dicombytesio":
            stream = DicomBytesIO(stream)
        elif stream_type == "dicom":
            stream = dummy_dicom_factory(stream)
        elif stream_type != "bytes":
            raise ValueError(stream_type)
        return stream

    return func


class TestBasicOffsetTable:
    @pytest.mark.parametrize(
        "offsets",
        [
            [0, 1000],
            [0, 1000, 2000],
        ],
    )
    def test_from_offsets(self, offsets):
        bot = BasicOffsetTable.from_offsets(offsets)
        assert len(bot) == BYTES_PER_LONG * len(offsets)
        assert list(bot) == offsets

    def test_default(self):
        bot = BasicOffsetTable.default()
        assert bot.num_frames == 0
        assert len(bot) == 0
        assert bot.total_length == 2 * BYTES_PER_LONG
        assert not list(bot)

    def test_repr(self):
        bot = BasicOffsetTable.default()
        s = repr(bot)
        assert isinstance(s, str)
        assert f"total_length={2 * BYTES_PER_LONG}" in s
        assert "frames=0" in s

    @pytest.mark.parametrize(
        "offsets",
        [
            [0, 1000],
            [0, 1000, 2000],
        ],
    )
    def test_from_stream(self, offsets, stream_type_factory):
        # use from_offsets to simplify setup
        src = BasicOffsetTable.from_offsets(offsets)
        needs_trimming = b"".join([b"\x00"] * 4)
        stream = b"".join([src.buffer, needs_trimming])
        stream = stream_type_factory(stream)
        bot = BasicOffsetTable.from_stream(stream)
        assert bot == src

    @pytest.mark.parametrize(
        "offsets",
        [
            [0, 1000],
            [0, 1000, 2000],
        ],
    )
    def test_total_length(self, offsets):
        bot = BasicOffsetTable.from_offsets(offsets)
        assert bot.total_length == BYTES_PER_LONG * (len(offsets) + 2)

    @pytest.mark.parametrize(
        "offsets",
        [
            [0, 1000],
            [0, 1000, 2000],
        ],
    )
    def test_num_frames(self, offsets):
        bot = BasicOffsetTable.from_offsets(offsets)
        assert bot.num_frames == len(offsets)

    @pytest.mark.parametrize(
        "offsets, exp",
        [
            pytest.param([0, 1000], True),
            pytest.param([0, 1000, 2000], True),
            pytest.param([0, 0, 1000], False),
            pytest.param([0, 1000, 1000], False),
            pytest.param([0, 1000, 100], False),
        ],
    )
    def test_is_valid(self, offsets, exp):
        bot = BasicOffsetTable.from_offsets(offsets)
        assert bot.is_valid == exp

    def test_is_valid_mod_length(self):
        bot = BasicOffsetTable.from_offsets([0, 1000])
        bot.fp.seek(LENGTH)
        bot.fp.write_UL(val=3)  # type: ignore
        assert not bot.is_valid

    @pytest.mark.parametrize(
        "offsets",
        [
            [0, 1000],
            [0, 1000, 2000],
        ],
    )
    def test_remove_from(self, offsets, stream_type_factory):
        src = BasicOffsetTable.from_offsets(offsets)
        extras = b"".join([b"\x00"] * 4)
        stream = b"".join([src.buffer, extras])
        stream = stream_type_factory(stream)
        result = BasicOffsetTable.remove_from(stream)
        assert result == extras

    @pytest.mark.parametrize(
        "offsets",
        [
            [0, 1000],
            [0, 1000, 2000],
        ],
    )
    def test_prepend_to(self, offsets, stream_type_factory):
        bot = BasicOffsetTable.from_offsets(offsets)
        extras = b"".join([b"\x00"] * 4)
        stream = stream_type_factory(extras)
        result = bot.prepend_to(stream)
        assert BasicOffsetTable.from_stream(result) == bot
        assert len(result) == bot.total_length + len(extras)

    @pytest.mark.parametrize(
        "offsets,val,exp",
        [
            pytest.param([0, 1000], 0, True),
            pytest.param([0, 1000, 2000], 2000, True),
            pytest.param([0, 1000], -1, False),
            pytest.param([0, 1000, 2000], 3000, False),
        ],
    )
    def test_contains(self, offsets, val, exp):
        bot = BasicOffsetTable.from_offsets(offsets)
        assert (val in bot) == exp



#!/usr/bin/env python
# -*- coding: utf-8 -*-
import runpy
import sys


def test_dicomphi(dicom_folder, tmp_path):
    sys.argv = [sys.argv[0], str(tmp_path)]
    runpy.run_module("dicom_utils.cli.dicomphi", run_name="__main__", alter_sys=True)



#!/usr/bin/env python
# -*- coding: utf-8 -*-

import numpy as np
import pytest

import dicom_utils
from dicom_utils import KeepVolume, RandomSlice, ReduceVolume, SliceAtLocation, UniformSample


class TestKeepVolume:
    def test_array(self):
        x = np.random.rand(10, 10)
        sampler = KeepVolume()
        result = sampler(x)
        assert (x == result).all()

    def test_dicom(self, dicom_object_3d, transfer_syntax):
        N = 8
        dcm = dicom_object_3d(N, syntax=transfer_syntax)
        sampler = KeepVolume()
        result = sampler(dcm)
        assert type(result) == type(dcm)
        assert result.NumberOfFrames == dcm.NumberOfFrames
        assert (result.pixel_array == dcm.pixel_array).all()


class TestSliceAtLocation:
    @pytest.mark.parametrize(
        "center,before,after,stride,index",
        [
            pytest.param(5, 0, 0, 1, lambda a: a[5]),
            pytest.param(5, 1, 0, 1, lambda a: a[4:6]),
            pytest.param(5, 1, 1, 1, lambda a: a[4:7]),
            pytest.param(5, 0, 0, 2, lambda a: a[5]),
            pytest.param(5, 2, 2, 2, lambda a: a[3:8:2]),
        ],
    )
    def test_array(self, center, before, after, stride, index):
        x = np.random.rand(10, 10)
        sampler = SliceAtLocation(center, before, after, stride)
        result = sampler(x)
        assert type(x) == type(result)
        assert (index(x) == result).all()

    @pytest.mark.parametrize(
        "center,before,after,stride",
        [
            pytest.param(5, 0, 0, 1),
            pytest.param(5, 1, 0, 1),
            pytest.param(5, 1, 1, 1),
            pytest.param(5, 0, 0, 2),
            pytest.param(5, 2, 2, 2),
        ],
    )
    def test_dicom(self, dicom_object_3d, center, before, after, stride, transfer_syntax):
        N = 8
        dcm = dicom_object_3d(N, syntax=transfer_syntax)

        sampler = SliceAtLocation(center, before, after, stride)
        result = sampler(dcm)
        assert dcm.NumberOfFrames == N, "the input dicom object was modified"
        assert type(result) == type(dcm)
        assert (sampler(dcm.pixel_array) == result.pixel_array).all()


class TestUniformSample:
    @pytest.mark.parametrize(
        "size,amount,method,index",
        [
            pytest.param(10, 2, "stride", lambda a: a[::2]),
            pytest.param(12, 2, "stride", lambda a: a[::2]),
            pytest.param(10, 4, "count", lambda a: a[:: 10 // 4]),
            pytest.param(12, 4, "count", lambda a: a[:: 12 // 4]),
            pytest.param(2, 4, "count", lambda a: a),
            pytest.param(2, 1, "stride", lambda a: a),
        ],
    )
    def test_array(self, size, amount, method, index):
        x = np.random.rand(size, 10)
        sampler = UniformSample(amount, method)
        result = sampler(x)
        assert type(x) == type(result)
        expected = index(x)
        try:
            assert (expected == result).all()
        except Exception:
            assert expected == result

    @pytest.mark.parametrize(
        "amount,method",
        [
            pytest.param(2, "stride"),
            pytest.param(2, "stride"),
            pytest.param(4, "count"),
            pytest.param(4, "count"),
            pytest.param(4, "count"),
            pytest.param(1, "stride"),
        ],
    )
    def test_dicom(self, dicom_object_3d, amount, method, transfer_syntax):
        N = 8
        dcm = dicom_object_3d(N, syntax=transfer_syntax)

        sampler = UniformSample(amount, method)
        result = sampler(dcm)
        assert dcm.NumberOfFrames == N, "the input dicom object was modified"
        assert type(result) == type(dcm)
        assert (sampler(dcm.pixel_array) == result.pixel_array).all()


class TestReduceVolume:
    def test_dicom(self, dicom_object_3d, transfer_syntax):
        N = 8
        dcm = dicom_object_3d(N, syntax=transfer_syntax)
        sampler = ReduceVolume()
        expected = np.max(dcm.pixel_array, axis=0)
        result = sampler(dcm)
        assert type(result) == type(dcm)
        assert result.NumberOfFrames == 1
        assert (result.pixel_array == expected).all()

    @pytest.mark.parametrize("output_frames", [1, 2, 4, 8])
    def test_multi_frame(self, dicom_object_3d, output_frames):
        N = 8
        dcm = dicom_object_3d(N)
        sampler = ReduceVolume(output_frames=output_frames)
        expected = dcm.pixel_array.reshape(output_frames, N // output_frames, *dcm.pixel_array.shape[1:]).max(axis=1)
        result = sampler(dcm)
        assert type(result) == type(dcm)
        assert (result.pixel_array == expected).all()
        assert result.NumberOfFrames == output_frames

    def test_multi_frame_noop(self, dicom_object_3d):
        N = 8
        dcm = dicom_object_3d(N)
        sampler = ReduceVolume(output_frames=N)
        result = sampler(dcm)
        assert type(result) == type(dcm)
        assert result.NumberOfFrames == N
        assert (result.pixel_array == dcm.pixel_array).all()

    @pytest.mark.parametrize(
        "skip,output_frames",
        [
            (0, 1),
            (2, 3),
            # In this case we widen the range of the output frames to include the edge frames
            (4, 3),
            # In this case we can't satisfy 9 output frames from 8 total frames
            pytest.param(4, 9, marks=pytest.mark.xfail(raises=RuntimeError, strict=True)),
        ],
    )
    def test_skip_edge_frames(self, dicom_object_3d, skip, output_frames):
        N = 8
        dcm = dicom_object_3d(N)
        sampler = ReduceVolume(output_frames=output_frames, skip_edge_frames=skip)
        result = sampler(dcm)
        assert type(result) == type(dcm)
        assert result.NumberOfFrames == output_frames

    def test_skip_edge_frames_noop_reduction(self, dicom_object_3d):
        N = 8
        dcm = dicom_object_3d(N)
        sampler = ReduceVolume(output_frames=N - 4, skip_edge_frames=2)
        expected = dcm.pixel_array[2:-2]
        result = sampler(dcm)
        assert type(result) == type(dcm)
        assert result.NumberOfFrames == N - 4
        assert (result.pixel_array == expected).all()

    @pytest.mark.parametrize("use_nvjpeg", [False, None])
    def test_decompress(self, mocker, use_nvjpeg, dicom_object_3d, transfer_syntax):
        spy = mocker.spy(dicom_utils.dicom, "decompress")
        N = 8
        dcm = dicom_object_3d(N, syntax=transfer_syntax)
        sampler = ReduceVolume()
        sampler(dcm, use_nvjpeg=use_nvjpeg)
        spy.assert_called()
        for call in spy.mock_calls:
            assert call.kwargs["use_nvjpeg"] == use_nvjpeg


class TestRandomSlice:
    def test_array(self):
        x = np.random.rand(10, 10)
        sampler = RandomSlice()
        result = sampler(x)
        assert result.shape == (1, 10)

    def test_dicom(self, dicom_object_3d, transfer_syntax):
        N = 8
        dcm = dicom_object_3d(N, syntax=transfer_syntax)
        sampler = RandomSlice()
        result = sampler(dcm)
        assert type(result) == type(dcm)
        assert result.NumberOfFrames == 1

    def test_seed(self):
        x = np.random.rand(10, 10)
        sampler1 = RandomSlice(seed=42)
        sampler2 = RandomSlice(seed=42)
        sampler3 = RandomSlice(seed=0)
        result1 = sampler1(x)
        result2 = sampler2(x)
        result3 = sampler3(x)
        assert (result1 == result2).all()
        assert not (result1 == result3).all()



#!/usr/bin/env python
# -*- coding: utf-8 -*-
from copy import copy
from functools import wraps
from inspect import signature
from struct import pack
from typing import Any, Callable, ClassVar, Final, Iterator, Sequence, TypeVar, Union, cast

from pydicom.filebase import DicomBytesIO

from .types import Dicom


StreamType = Union[bytes, DicomBytesIO, Dicom]
T = TypeVar("T")

START: Final = 0
PREFIX: Final = 0
BYTES_PER_LONG: Final = 4
LENGTH: Final = 4
OFFSETS: Final = 8


def _prepare_stream(stream: StreamType) -> DicomBytesIO:
    if isinstance(stream, Dicom):
        if not hasattr(stream, "PixelData"):
            raise ValueError("Input DICOM object does not have PixelData")
        stream = DicomBytesIO(stream.PixelData)
    elif isinstance(stream, bytes):
        stream = DicomBytesIO(stream)
    elif not isinstance(stream, DicomBytesIO):
        raise TypeError(type(stream))
    stream.is_little_endian = True
    stream.seek(0)
    return stream


# TODO: annotated with ParamSpec once Python 3.10 is well adopted
def prepare_stream(f: Callable[..., T]) -> Callable[..., T]:
    r"""Decorator to prepare args annotated as DicomBytesIO from a StreamType.
    Automatically converts the byte stream and seeks to the start of stream.
    """
    param_dict = signature(f).parameters
    param_list = list(param_dict)

    @wraps(f)
    def wrapper(*args, **kwargs) -> T:
        args = tuple(
            a if param_dict[param_list[i]].annotation != DicomBytesIO else _prepare_stream(a)
            for i, a in enumerate(args)
        )
        kwargs = {k: v if param_dict[k].annotation != DicomBytesIO else _prepare_stream(v) for k, v in kwargs.items()}
        return f(*args, **kwargs)

    return wrapper


def restore_pointer(f: Callable[..., T]) -> Callable[..., T]:
    r"""Decorator to restore the stream position after a function call"""

    @wraps(f)
    def wrapper(self: "BasicOffsetTable", *args, **kwargs) -> T:
        pos = self.fp.tell()
        result = f(self, *args, **kwargs)
        self.fp.seek(pos)
        return result

    return wrapper


class BasicOffsetTable:
    r"""Helper class for manipulating the Basic Offset Table"""

    PREFIX: ClassVar[bytes] = b"\xFE\xFF\x00\xE0"
    DELIM: ClassVar[bytes] = b"\xFF\xFE\xE0\xDD"

    def __init__(self, fp: DicomBytesIO):
        if not isinstance(fp, DicomBytesIO):
            raise TypeError(type(fp))
        self.fp = copy(fp)
        self.fp.is_little_endian = True

        if not self.is_present(self.fp):
            raise ValueError("A basic offset table was not present")

        # trim the input stream to just the BOT
        size = self.total_length
        self.fp.seek(START)
        trimmed = self.fp.read(size)
        self.fp = DicomBytesIO(trimmed)
        self.fp.is_little_endian = True
        assert self.is_present(self.fp)

    def __eq__(self, other: Any) -> bool:
        if not isinstance(other, BasicOffsetTable):
            return False
        return self.buffer == other.buffer

    def __repr__(self) -> str:
        return f"{self.__class__.__name__}(total_length={self.total_length}, frames={self.num_frames})"

    @restore_pointer
    def __len__(self) -> int:
        r"""The value of the length component of the BOT"""
        self.fp.seek(LENGTH)
        return self.fp.read_UL()

    @property
    def buffer(self) -> bytes:
        r"""The raw bytes of this BOT"""
        # type errors when using restore_pointer, so handle it manually here
        pos = self.fp.tell()
        self.fp.seek(START)
        result = self.fp.read()
        self.fp.seek(pos)
        return result

    @property
    def total_length(self) -> int:
        r"""The total number of bytes in this BOT"""
        return len(self) + 2 * BYTES_PER_LONG

    @property
    def num_frames(self) -> int:
        r"""The total number of frames in this BOT"""
        return len(self) // 4

    @restore_pointer
    def __iter__(self) -> Iterator[int]:
        r"""Iterator over frame offsets"""
        steps = len(self) // 4
        self.fp.seek(OFFSETS)
        for i in range(steps):
            yield self.fp.read_UL()

    @property
    def is_valid(self) -> bool:
        r"""Checks if this basic offset table is valid.
        A basic offset table should have a length divisible by 4, and the frame
        offsets should be monotonically increasing.
        """
        return len(self) % BYTES_PER_LONG == 0 and list(self) == sorted(set(self))

    @classmethod
    @prepare_stream
    def is_present(cls, stream: DicomBytesIO) -> bool:
        r"""Checks if a stream contains a BOT"""
        pos = stream.tell()
        stream.seek(PREFIX)
        prefix = stream.read(BYTES_PER_LONG)
        stream.seek(pos)
        return prefix == cls.PREFIX

    @classmethod
    def from_offsets(cls, offsets: Sequence[int]) -> "BasicOffsetTable":
        r"""Create a BOT from a sequence of frame offsets"""
        stream = bytearray()
        # prefix and length
        stream.extend(cls.PREFIX)
        stream.extend(pack("<I", BYTES_PER_LONG * len(offsets)))
        for offset in offsets:
            stream.extend(pack("<I", offset))
        return cls.from_stream(bytes(stream))

    @classmethod
    @prepare_stream
    def from_stream(cls, stream: DicomBytesIO) -> "BasicOffsetTable":
        r"""Create a BOT from a byte stream"""
        stream.seek(START)
        return cls(stream)

    @classmethod
    def default(cls) -> "BasicOffsetTable":
        r"""Create a default (empty) BOT"""
        data = b"".join([cls.PREFIX, b"\x00\x00\x00\x00"])
        return cls.from_stream(data)

    @classmethod
    @prepare_stream
    def remove_from(cls, stream: DicomBytesIO) -> bytes:
        r"""Removes the BOT from a stream"""
        pos = stream.tell()
        if cls.is_present(stream):
            bot = cls.from_stream(stream)
            stream.seek(bot.total_length)
        else:
            stream.seek(START)
        result = stream.read()
        stream.seek(pos)
        return result

    @prepare_stream
    def prepend_to(self, stream: DicomBytesIO) -> bytes:
        r"""Prepends this BOT to the start of a stream"""
        stream.seek(START)
        buffer = cast(bytes, self.buffer)
        return b"".join([buffer, stream.read()])



#!/usr/bin/env python
# -*- coding: utf-8 -*-
from pathlib import Path

import pytest
import tqdm_multiprocessing

from dicom_utils.cli.organize import organize
from dicom_utils.dicom_factory import CompleteMammographyStudyFactory, DicomFactory


class TestSymlinkPipeline:
    @pytest.mark.parametrize("implants", [False, True])
    @pytest.mark.parametrize("spot", [False, True])
    def test_case_dir_structure(self, tmp_path, implants, spot):
        types = ("ffdm", "tomo", "synth", "ultrasound")
        paths = []
        seen_sop_uids = []
        for i in range(num_cases := 3):
            factory = CompleteMammographyStudyFactory(
                types=types,
                spot_compression=spot,
                implants=implants,
                seed=i,
            )
            case_dir = Path(tmp_path, f"Original-{i}")
            case_dir.mkdir(parents=True)
            dicoms = factory(StudyInstanceUID=f"study-{i}")
            outputs = DicomFactory.save_dicoms(case_dir, dicoms)
            paths.append(outputs)
            seen_sop_uids = seen_sop_uids + [d.SOPInstanceUID for d in dicoms]

        # non-deterministic test failures will happen if the UIDs aren't unique
        assert len(seen_sop_uids) == len(set(seen_sop_uids))

        dest = Path(tmp_path, "symlinks")
        dest.mkdir()
        result = organize(tmp_path, dest, threads=False, use_bar=False, jobs=4)
        for _, results in result.items():
            recs = {k: [r.path.relative_to(dest) for r in v] for k, v in results.items()}
            assert len(recs) == num_cases
            assert all(v for v in recs.values())

    @pytest.mark.parametrize("threads", [True, False])
    @pytest.mark.parametrize("jobs", [4, 8])
    def test_forward_jobs_and_threads(self, tmp_path, mocker, threads, jobs):
        spy = mocker.spy(tqdm_multiprocessing.ConcurrentMapper, "__init__")

        paths = []
        for i in range(num_cases := 3):
            factory = CompleteMammographyStudyFactory(
                seed=i,
            )
            case_dir = Path(tmp_path, f"Original-{i}")
            case_dir.mkdir(parents=True)
            dicoms = factory(StudyInstanceUID=f"study-{i}")
            outputs = DicomFactory.save_dicoms(case_dir, dicoms)
            paths.append(outputs)

        dest = Path(tmp_path, "symlinks")
        dest.mkdir()
        # set use_bar = True to help debugging
        timeout = 2
        organize(tmp_path, dest, use_bar=False, jobs=jobs, threads=threads, timeout=timeout)
        assert spy.call_count == 3
        for i, call in enumerate(spy.call_args_list):
            if i not in (1, 2):
                assert call.args[0].threads == threads, f"call {i} failed"
            else:
                # Grouper has thread override because of deadlocks
                assert call.args[0].threads, f"call {i} failed"
            assert call.args[0].jobs == jobs, f"call {i} failed"
            assert call.args[0].timeout == timeout, f"call {i} failed"



#!/usr/bin/env python
# -*- coding: utf-8 -*-
import argparse
import sys
from argparse import ArgumentParser
from io import BytesIO
from pathlib import Path
from typing import List, Optional

import matplotlib.pyplot as plt
import numpy as np
from PIL import Image

from ..dicom import path_to_dicoms
from ..types import Dicom
from ..visualize import DicomImage, chw_to_hwc, dcms_to_annotated_images, to_collage
from ..volume import VOLUME_HANDLERS, KeepVolume, VolumeHandler


def get_parser(parser: ArgumentParser = ArgumentParser()) -> ArgumentParser:
    parser.add_argument("path", help="DICOM file or folder with files to convert")
    parser.add_argument("-o", "--output", help="output filepath. if directory, use filename from `file`")
    parser.add_argument("-d", "--downsample", help="downsample images by an integer factor", type=int)
    parser.add_argument(
        "-s", "--split", default=False, action="store_true", help="split multi-frame inputs into separate files"
    )
    parser.add_argument("-f", "--fps", default=5, type=int, help="framerate for animated outputs")
    parser.add_argument("-q", "--quality", default=95, type=int, help="quality of outputs, from 1 to 100")
    parser.add_argument("--noblock", default=False, action="store_true", help="allow matplotlib to block")
    parser.add_argument("-b", "--bytes", default=False, action="store_true", help="output a png image as a byte stream")
    parser.add_argument(
        "-v",
        "--volume-handler",
        default="keep",
        choices=VOLUME_HANDLERS.available_keys(),
        help="volume handler for 3D inputs",
    )
    return parser


class ImageOutput:
    r"""Base class that describes how to render a given collage."""

    def __init__(self, images: List[DicomImage], quality: int = 95, downsample: int = 1, **kwargs):
        self.quality = quality
        self.images = images
        self.data = to_collage([i.pixels[:, :, ::downsample, ::downsample] for i in self.images])
        if self.is_single_frame:
            self.data = self.data[0]

    @property
    def chw_data(self) -> np.ndarray:
        return self.data

    @property
    def hwc_data(self) -> np.ndarray:
        return chw_to_hwc(self.data)

    @property
    def is_single_frame(self) -> bool:
        return all(i.is_single_frame for i in self.images)

    def to_bytes(self) -> bytes:
        img = Image.fromarray(self.hwc_data)
        buf = BytesIO()
        img.save(buf, format="png", quality=self.quality)
        return buf.getvalue()

    def print_bytes(self) -> None:
        sys.stdout.buffer.write(self.to_bytes())

    def save(self, path: Path) -> None:
        img = Image.fromarray(self.hwc_data)
        img.save(str(path), quality=self.quality)

    def show(self, block: bool = True, **kwargs) -> None:
        print("Showing image")
        plt.imshow(self.hwc_data, **kwargs)
        plt.show(block=block)

    @staticmethod
    def from_dicom_images(images: List[DicomImage], split: bool = False, **kwargs) -> "ImageOutput":
        if all(i.is_single_frame for i in images):
            return ImageOutput(images, **kwargs)
        elif split:
            return SplitImageOutput(images, **kwargs)
        else:
            return AnimatedImageOutput(images, **kwargs)


class SplitImageOutput(ImageOutput):
    r"""Renderer for 3D data that splits the 3D volume into multiple 2D frames"""

    def to_bytes(self) -> bytes:
        buf = BytesIO()
        for i, frame in enumerate(self.data):
            img = Image.fromarray(chw_to_hwc(frame))
            img.save(buf, format="png", quality=self.quality)
        return buf.getvalue()

    def save(self, path: Path) -> None:
        subdir = Path(path.with_suffix(""))
        subdir.mkdir(exist_ok=True)
        for i, frame in enumerate(self.data):
            path = Path(subdir, Path(f"{i}.png"))
            img = Image.fromarray(chw_to_hwc(frame))
            img.save(str(path), quality=self.quality)

    def show(self, **kwargs) -> None:
        for i, frame in enumerate(self.data):
            print(f"Showing frame {i}/{len(self.data)}")
            plt.imshow(chw_to_hwc(frame), cmap="gray")
            plt.show(block=True)


class AnimatedImageOutput(ImageOutput):
    r"""Renderer for 3D data that converts the volume to a 2D animation"""

    def __init__(self, images: List[DicomImage], quality: int = 95, downsample: int = 1, fps: int = 5):
        super().__init__(images, quality, downsample)
        self.fps = fps

    def _prepare_frames(self) -> List[Image.Image]:
        frames = [Image.fromarray(chw_to_hwc(frame)) for frame in self.data]
        return frames

    @property
    def duration_ms(self) -> float:
        return len(self.data) / (self.fps * 1000)

    def to_bytes(self) -> bytes:
        buf = BytesIO()
        frames = self._prepare_frames()
        frames[0].save(
            buf,
            save_all=True,
            append_images=frames[1:],
            duration=self.duration_ms,
            quality=self.quality,
            format="gif",
        )
        return buf.getvalue()

    def save(self, path: Path) -> None:
        frames = self._prepare_frames()
        path = path.with_suffix(".gif")
        frames[0].save(path, save_all=True, append_images=frames[1:], duration=self.duration_ms, quality=self.quality)

    def show(self, *args, **kwargs) -> None:
        raise NotImplementedError(
            "Showing an animated image is not yet implemented. "
            "Please save to an output GIF file or utilize piped byte stream"
        )


def dicoms_to_graphic(
    dcms: List[Dicom],
    dest: Optional[Path] = None,
    split: bool = False,
    fps: int = 5,
    quality: int = 95,
    block: bool = True,
    as_bytes: bool = False,
    downsample: int = 1,
    volume_handler: VolumeHandler = KeepVolume(),
    **kwargs,
) -> None:
    images = dcms_to_annotated_images(dcms, volume_handler=volume_handler, **kwargs)

    handler = ImageOutput.from_dicom_images(
        images,
        split,
        quality=quality,
        downsample=downsample,
        fps=fps,
    )

    if as_bytes:
        handler.print_bytes()
    elif dest is not None:
        handler.save(dest)
    else:
        handler.show(block)


def main(args: argparse.Namespace) -> None:
    path = Path(args.path)
    dest = Path(args.output) if args.output is not None else None

    # handle case where output path is a dir
    if dest is not None and dest.is_dir():
        dest = Path(dest, path.stem).with_suffix(".png")

    volume_handler = VOLUME_HANDLERS.get(args.volume_handler).instantiate_with_metadata().fn
    assert isinstance(volume_handler, VolumeHandler)

    dcms = list(path_to_dicoms(path))
    dicoms_to_graphic(
        dcms,
        dest,
        args.split,
        args.fps,
        args.quality,
        not args.noblock,
        args.bytes,
        args.downsample,
        volume_handler,
    )


def entrypoint():
    parser = get_parser()
    main(parser.parse_args())


if __name__ == "__main__":
    entrypoint()



#!/usr/bin/env python
# -*- coding: utf-8 -*-
import importlib
import inspect
import json
import logging
import sys
from abc import ABC
from collections import defaultdict
from dataclasses import dataclass, field, fields, replace
from functools import cached_property, partial
from io import BytesIO, IOBase
from itertools import chain
from os import PathLike
from pathlib import Path, PosixPath
from statistics import mode
from typing import (
    Any,
    Dict,
    Final,
    Hashable,
    Iterable,
    Iterator,
    List,
    Optional,
    Set,
    Tuple,
    Type,
    TypeVar,
    Union,
    cast,
)

import pydicom
from pydicom import DataElement, Dataset, Sequence
from pydicom.multival import MultiValue
from pydicom.uid import SecondaryCaptureImageStorage


logger = logging.getLogger(__name__)


from registry import Registry

from ..dicom import Dicom
from ..tags import Tag

# Type checking fails when dataclass attr name matches a type alias.
# Import types under a different alias
from ..types import DicomKeyError, DicomValueError
from ..types import ImageType as IT
from ..types import Laterality, MammogramType, MammogramView
from ..types import PhotometricInterpretation as PI
from ..types import PixelSpacing as PixelSpacingClass
from ..types import ViewPosition, get_value, iterate_view_modifier_codes
from .helpers import SOPUID, ImageUID, SeriesUID, StudyUID
from .helpers import TransferSyntaxUID as TSUID
from .protocols import (
    SupportsDataSetInfo,
    SupportsGenerated,
    SupportsManufacturer,
    SupportsPatientAge,
    SupportsPatientID,
    SupportsSite,
    SupportsStudyDate,
    SupportsStudyUID,
    SupportsUID,
)


STANDARD_MAMMO_VIEWS: Final[Set[MammogramView]] = {
    MammogramView(Laterality.LEFT, ViewPosition.MLO),
    MammogramView(Laterality.RIGHT, ViewPosition.MLO),
    MammogramView(Laterality.LEFT, ViewPosition.CC),
    MammogramView(Laterality.RIGHT, ViewPosition.CC),
}


R = TypeVar("R", bound="FileRecord")
T = TypeVar("T")

RECORD_REGISTRY = Registry("records")
HELPER_REGISTRY = Registry("helpers")
SEP: Final[str] = "_"
DEFAULT_FILE_ID: Final = 1
FILE_ID_IDX: Final = -1


class StandardizedFilename(PosixPath):
    if sys.version_info >= (3, 12):

        def __init__(self, *args, **kwargs):
            path = PosixPath(*args, **kwargs)
            if SEP not in path.name:
                parts = str(path.with_suffix("")).split(SEP) + [str(DEFAULT_FILE_ID)]
                path = PosixPath(SEP.join(parts)).with_suffix(path.suffix)
            super().__init__(path)  # type: ignore

    else:

        def __new__(cls, *args, **kwargs):
            self = super().__new__(cls, *args, **kwargs)
            if SEP not in self.name:
                parts = self.tokenize() + [str(DEFAULT_FILE_ID)]
                self = StandardizedFilename(SEP.join(parts)).with_suffix(self.suffix)
            return self

    @property
    def prefix(self) -> str:
        return SEP.join(self.prefix_parts)

    @property
    def prefix_parts(self) -> List[str]:
        return self.tokenize()[:FILE_ID_IDX]

    @property
    def file_id(self) -> str:
        return self.tokenize()[FILE_ID_IDX]

    def with_file_id(self, val: Any) -> "StandardizedFilename":
        parts = self.tokenize()[:FILE_ID_IDX]
        parts.append(str(val))
        return StandardizedFilename(SEP.join(parts)).with_suffix(self.suffix)

    def with_prefix(self, *val: str) -> "StandardizedFilename":
        parts = list(val) + [self.file_id]
        return StandardizedFilename(SEP.join(parts)).with_file_id(self.file_id).with_suffix(self.suffix)

    def add_modifier(self, val: Any) -> "StandardizedFilename":
        parts = self.tokenize()
        parts.insert(-1, str(val))
        return StandardizedFilename(SEP.join(parts)).with_suffix(self.suffix)

    def tokenize(self) -> List[str]:
        return str(self.with_suffix("")).split(SEP)


@RECORD_REGISTRY(name="file")
@dataclass(frozen=True, order=True)
class FileRecord:
    path: Path = field(compare=True)

    def __post_init__(self):
        object.__setattr__(self, "path", Path(self.path))

    def __repr__(self) -> str:
        contents = [f"{name}={value}" for name, value in self.present_fields()]
        return f"{self.__class__.__name__}({', '.join(contents)})"

    def __hash__(self) -> int:
        return hash(self.path)

    def __eq__(self, other: Any) -> bool:
        return isinstance(other, type(self)) and self.path == other.path

    def replace(self: R, **kwargs) -> R:
        return replace(self, **kwargs)

    @property
    def is_symlink(self) -> bool:
        return self.path.is_symlink()

    @property
    def exists(self) -> bool:
        return self.path.is_file()

    @property
    def file_size(self) -> int:
        return self.path.stat().st_size

    @property
    def is_compressed(self) -> bool:
        return False

    @property
    def has_uid(self) -> bool:
        return bool(self.path)

    def get_uid(self) -> Hashable:
        return self.path.stem

    @classmethod
    def from_file(cls: Type[R], path: PathLike, helpers: Iterable["RecordHelper"] = []) -> R:
        path = Path(path)
        if not path.is_file():
            raise FileNotFoundError(path)
        result = cls(path)
        result = apply_helpers(path, result, helpers)
        return result

    def relative_to(self: R, target: Union[PathLike, "FileRecord"]) -> R:
        path = target.path if isinstance(target, FileRecord) else Path(target)
        # `path` is a parent of `self.path`
        if self.path.is_relative_to(path):
            return replace(self, path=self.path.relative_to(path))

        # `self.path` shares a parent with `path`
        path = path.absolute()
        self_path = self.path.absolute()
        paths_to_check = [path, *path.parents]
        for i, parent in enumerate(paths_to_check):
            if self_path.is_relative_to(parent):
                relpath = self_path.relative_to(parent)
                self_path = Path(*([".."] * i), relpath)
                break
        else:
            raise ValueError(f"Record path {self.path} is not relative to {path}")
        return replace(self, path=self_path)

    def shares_directory_with(self, other: "FileRecord") -> bool:
        return self.path.absolute().parent == other.path.absolute().parent

    def absolute(self: R) -> R:
        return replace(self, path=self.path.absolute())

    def present_fields(self) -> Iterator[Tuple[str, Any]]:
        for f in fields(self):
            value = getattr(self, f.name)
            if value != f.default:
                yield f.name, value

    def standardized_filename(self, file_id: Any = None) -> StandardizedFilename:
        file_id = str(file_id) if file_id is not None else str(self.get_uid() if self.has_uid else "1")
        file_id = file_id.replace(".", "-")
        return StandardizedFilename(self.path.name).with_file_id(file_id)

    @classmethod
    def read(
        cls,
        target: Union[PathLike, "FileRecord"],
        *args,
        helpers: Iterable["RecordHelper"] = [],
        **kwargs,
    ) -> IOBase:
        r"""Reads a DICOM file with optimized defaults for :class:`DicomFileRecord` creation.

        Args:
            path: Path to DICOM file to read

        Keyword Args:
            Overrides forwarded to :func:`pydicom.dcmread`
        """
        path = Path(target.path if isinstance(target, FileRecord) else target)
        if not path.is_file():
            raise FileNotFoundError(path)
        result = open(path, *args, **kwargs)
        result = apply_read_helpers(result, cls, helpers)
        return result

    def to_symlink(self: R, symlink_path: PathLike, overwrite: bool = False) -> R:
        r"""Create a symbolic link to the file referenced by this :class:`FileRecord`.
        The symbolic link will be relative to the location of the file referenced by this
        :class:`FileRecord`.

        Args:
            symlink_path:
                Filepath for the output symlink

            overwrite:
                If ``True``, and ``symlink_path`` is an existing symbolic link, overwrite it

        Returns:
            A new :class:`FileRecord` with ``path`` set to ``symlink_path``.
        """
        symlink_record = self.replace(path=symlink_path)
        symlink_record.path.parent.mkdir(exist_ok=True, parents=True)
        symlink_contents = Path(*self.relative_to(symlink_record).path.parts[1:])
        if overwrite:
            symlink_record.path.unlink(missing_ok=True)
        symlink_record.path.symlink_to(symlink_contents)

        resolved_path = symlink_record.path.resolve().absolute()
        real_path = self.path.resolve().absolute()
        assert resolved_path == real_path, f"{resolved_path} did not match {real_path}"
        return symlink_record

    def to_dict(self, file_id: Any = None) -> Dict[str, Any]:
        result: Dict[str, Any] = {}
        result["record_type"] = self.__class__.__name__
        module = inspect.getmodule(self.__class__)
        module = module.__name__ if module is not None else ""
        if module == "__main__":
            import warnings

            warnings.warn("TODO")
        result["record_type_import"] = module
        result["path"] = str(self.path)
        result["resolved_path"] = str(self.path.resolve())
        return result

    @staticmethod
    def from_dict(target: Union[Path, Dict[str, Any]]) -> "FileRecord":
        r"""Create a :class:`FileRecord` from a dictionary. The dictionary should have been
        created with :meth:`FileRecord.to_dict`. Keys ``record_type`` and ``record_type_import``
        are used to determine the class to instantiate.
        """
        if isinstance(target, Path):
            with open(target) as f:
                target = json.load(f)
        elif not isinstance(target, dict):
            raise TypeError(f"`target` should be a Path or dict, got {type(target)}")

        assert isinstance(target, dict)
        import_path = target["record_type_import"]
        class_name = target["record_type"]
        try:
            # get the class to be instantiated
            mod = importlib.import_module(import_path)
            cls: Type[FileRecord] = getattr(mod, class_name)
            # run classmethod for instantiation
            return cls._from_dict(target)

        except Exception as ex:
            logger.warn("Failed to create FileRecord from dict")
            logger.warn("Exception info", exc_info=ex)
            record_path = Path(target["path"])
            return FileRecord(record_path)

    @classmethod
    def _from_dict(cls: Type[R], target: Dict[str, Any]) -> R:
        # find dataclass attributes that exist in the dict
        kwargs: Dict[str, Any] = {}
        dest_fields = set(f.name for f in fields(cls))
        for k, v in target.items():
            if k in dest_fields:
                kwargs[k] = v
        record_path = Path(kwargs.pop("path"))
        rec = cls(record_path, **kwargs)
        return rec


# NOTE: record contents should follow this naming scheme:
#   * When a DICOM tag is read directly, attribute name should match tag name.
#     E.g. Tag StudyInstanceUID -> Attribute StudyInstanceUID
#   * When a one or more DICOM tags are read with additional parsing logic, attribute
#     name should differ from tag name. E.g. attribute `view_position` has advanced parsing
#     logic that reads over multiple tags
# TODO: find a way to functools.partial this dataclass decorator that works with type checker


@RECORD_REGISTRY(name="dicom", suffixes=[".dcm"])
@dataclass(frozen=True, order=False, eq=False)
class DicomFileRecord(
    FileRecord,
    SupportsStudyUID,
    SupportsStudyDate,
    SupportsManufacturer,
    SupportsUID,
    SupportsPatientID,
    SupportsPatientAge,
    SupportsGenerated,
    SupportsSite,
    SupportsDataSetInfo,
):
    r"""Data structure for storing critical information about a DICOM file.
    File IO operations on DICOMs can be expensive, so this class collects all
    required information in a single pass to avoid repeated file opening.
    """

    StudyInstanceUID: Optional[StudyUID] = None
    SeriesInstanceUID: Optional[SeriesUID] = None
    SOPInstanceUID: Optional[SOPUID] = None
    SOPClassUID: Optional[SOPUID] = None
    TransferSyntaxUID: Optional[TSUID] = None
    Modality: Optional[str] = None
    BodyPartExamined: Optional[str] = None
    PatientOrientation: Optional[List[str]] = None
    StudyDate: Optional[str] = None
    ContentDate: Optional[str] = None
    AcquisitionDate: Optional[str] = None
    SeriesDescription: Optional[str] = None
    StudyDescription: Optional[str] = None
    PatientName: Optional[str] = None
    PatientID: Optional[str] = None
    StudyID: Optional[str] = None
    Manufacturer: Optional[str] = None
    ManufacturerModelName: Optional[str] = None
    ManufacturerModelNumber: Optional[str] = None
    PresentationIntentType: Optional[str] = None
    PatientAge: Optional[str] = None
    PatientBirthDate: Optional[str] = None
    BodyPartThickness: Optional[str] = None
    InstitutionAddress: Optional[str] = None
    InstitutionName: Optional[str] = None
    TreatmentSite: Optional[str] = None
    DataSetName: Optional[str] = None
    DataSetSource: Optional[str] = None
    DataSetDescription: Optional[str] = None
    DataSetType: Optional[str] = None
    DataSetSubtype: Optional[str] = None
    PerformedProcedureStepDescription: Optional[str] = None
    AccessionNumber: Optional[str] = None

    generated: bool = False
    is_cad: bool = False

    def __eq__(self, other: Any) -> bool:
        if not isinstance(other, type(self)):
            return False

        if self.SOPInstanceUID and other.SOPInstanceUID:
            return self.same_uid_as(other)
        return self.path == other.path

    def __lt__(self, other: FileRecord) -> bool:
        if not isinstance(other, DicomFileRecord):
            return FileRecord(self.path) < other
        return str(self.SOPInstanceUID) < str(other.SOPInstanceUID)

    def __gt__(self, other: FileRecord) -> bool:
        if not isinstance(other, DicomFileRecord):
            return FileRecord(self.path) > other
        return str(self.SOPInstanceUID) > str(other.SOPInstanceUID)

    def __le__(self, other: FileRecord) -> bool:
        if not isinstance(other, DicomFileRecord):
            return FileRecord(self.path) <= other
        return str(self.SOPInstanceUID) <= str(other.SOPInstanceUID)

    def __ge__(self, other: FileRecord) -> bool:
        if not isinstance(other, DicomFileRecord):
            return FileRecord(self.path) >= other
        return str(self.SOPInstanceUID) >= str(other.SOPInstanceUID)

    def __hash__(self) -> int:
        if self.has_uid:
            return hash(self.get_uid(prefer_sop=True))
        return hash(self.path)

    def __iter__(self) -> Iterator[Tuple[Tag, Any]]:
        for f in fields(self):
            name = f.name
            value = getattr(self, name)
            if (tag := getattr(Tag, f.name, None)) is not None and value is not None:
                yield tag, value

    @cached_property
    def is_for_processing(self) -> bool:
        return (self.PresentationIntentType or "").lower() == "for processing"

    @property
    def is_compressed(self) -> bool:
        return False

    @property
    def is_secondary_capture(self) -> bool:
        return (self.SOPClassUID or "") == SecondaryCaptureImageStorage

    @cached_property
    def is_diagnostic(self) -> bool:
        if "diag" in (self.StudyDescription or "").lower():
            return True
        return False

    @cached_property
    def is_screening(self) -> bool:
        if "screening" in (self.StudyDescription or "").lower():
            return True
        return False

    @property
    def is_pr_file(self) -> bool:
        return self.Modality == "PR"

    @property
    def year(self) -> Optional[int]:
        r"""Extracts a year from ``Tag.StudyDate``.

        Returns:
            First 4 digits of ``Tag.StudyDate`` as an int, or None if a year could not be parsed
        """
        if self.StudyDate and len(self.StudyDate) > 4:
            try:
                return int(self.StudyDate[:4])
            except Exception:
                pass
        return None

    def to_dict(self) -> Dict[str, Any]:
        result = super().to_dict()
        for tag, value in self:
            if not isinstance(value, Sequence):
                result[tag.name] = value
        result["secondary_capture"] = self.is_secondary_capture
        result["diagnostic"] = self.is_diagnostic
        result["screening"] = self.is_screening
        result["is_cad"] = self.is_cad
        result["generated"] = self.generated
        return result

    @classmethod
    def from_file(
        cls,
        path: PathLike,
        helpers: Iterable["RecordHelper"] = [],
        overrides: Dict[str, Any] = {},
        **kwargs,
    ) -> "DicomFileRecord":
        r"""Creates a :class:`DicomFileRecord` from a DICOM file.

        Args:
            path: Path to DICOM file
            helpers: List of :class:`RecordHelper` objects to use for parsing
            overrides: Dictionary of overrides to apply to the record

        Keyword Args:
            Forwarded to :func:`DicomFileRecord.read`
        """
        path = Path(path)
        if not path.is_file():
            raise FileNotFoundError(path)
        with cls.read(path, helpers=helpers, **kwargs) as dcm:
            result = cls.from_dicom(path, dcm, **overrides)
        result = apply_helpers(path, result, helpers)
        return result

    @classmethod
    def from_dicom(
        cls,
        path: PathLike,
        dcm: Dicom,
        helpers: Iterable["RecordHelper"] = [],
        **overrides,
    ) -> "DicomFileRecord":
        r"""Creates a :class:`DicomFileRecord` from a DICOM file.

        Args:
            path: Path to DICOM file (needed to set ``path`` attribute)
            dcm: Dicom file object
            helpers: List of :class:`RecordHelper` objects to use for parsing

        Keyword Args:
            Overrides to apply to the record
        """
        path = Path(path)
        values = {tag.name: get_value(dcm, tag, None, try_file_meta=True) for tag in cls.get_required_tags()}
        values.update(overrides)

        # pop any values that arent part of the DicomFileRecord constructor, such as intermediate tags
        keyword_values = set(f.name for f in fields(cls))
        values = {k: v for k, v in values.items() if k in keyword_values}

        rec = cls(
            path.absolute(),
            **values,
        )
        rec = apply_helpers(path.absolute(), rec, helpers)
        return rec

    @property
    def has_uid(self) -> bool:
        r"""Tests if the record has a SeriesInstanceUID or SOPInstanceUID"""
        return bool(self.SeriesInstanceUID or self.SOPInstanceUID)

    def get_uid(self, prefer_sop: bool = True) -> ImageUID:
        r"""Gets an image level UID. The UID will be chosen from SeriesInstanceUID and SOPInstanceUID,
        with preference as specified in ``prefer_sop``.
        """
        if not self.has_uid:
            raise AttributeError("DicomFileRecord has no UID")
        if prefer_sop:
            result = self.SOPInstanceUID or self.SeriesInstanceUID
        else:
            result = self.SeriesInstanceUID or self.SOPInstanceUID
        assert result is not None
        return result

    @classmethod
    def read(cls, path: PathLike, *args, helpers: Iterable["RecordHelper"] = [], **kwargs) -> Dicom:
        r"""Reads a DICOM file with optimized defaults for :class:`DicomFileRecord` creation.

        Args:
            path: Path to DICOM file to read

        Keyword Args:
            Overrides forwarded to :func:`pydicom.dcmread`
        """
        kwargs.setdefault("stop_before_pixels", True)
        kwargs.setdefault("specific_tags", cls.get_required_tags())
        stream = cast(BytesIO, FileRecord.read(path, "rb"))
        result = pydicom.dcmread(stream, *args, **kwargs)
        result = apply_read_helpers(result, cls, helpers)
        return result

    def standardized_filename(self, file_id: Optional[str] = None) -> StandardizedFilename:
        path = super().standardized_filename(file_id)
        path = path.with_prefix(self.Modality.lower() if self.Modality else "unknown")
        if self.is_secondary_capture:
            path = path.add_modifier("secondary")
        if self.is_for_processing:
            path = path.add_modifier("proc")
        if self.is_cad:
            path = path.add_modifier("cad")
        return path

    @classmethod
    def get_required_tags(cls) -> Set[Tag]:
        return {getattr(Tag, field.name) for field in fields(cls) if hasattr(Tag, field.name)}


@RECORD_REGISTRY(name="dicom-image", suffixes=[".dcm"])
@dataclass(frozen=True, order=False, eq=False)
class DicomImageFileRecord(DicomFileRecord):
    r"""Data structure for storing critical information about a DICOM file.
    File IO operations on DICOMs can be expensive, so this class collects all
    required information in a single pass to avoid repeated file opening.
    """

    TransferSyntaxUID: Optional[TSUID] = None

    Rows: Optional[int] = None
    Columns: Optional[int] = None
    NumberOfFrames: Optional[int] = None
    PhotometricInterpretation: Optional[PI] = None
    ImageType: Optional[IT] = None
    BitsStored: Optional[int] = None
    PixelSpacing: Optional[Union[str, MultiValue]] = None
    ImagerPixelSpacing: Optional[Union[str, MultiValue]] = None
    ViewCodeSequence: Optional[Dataset] = None
    ViewModifierCodeSequence: Optional[Dataset] = None
    ViewPosition: Optional[str] = None

    @property
    def is_valid_image(self) -> bool:
        return bool(self.Rows and self.Columns and self.PhotometricInterpretation)

    @property
    def is_compressed(self) -> bool:
        return bool(self.TransferSyntaxUID) and self.TransferSyntaxUID.is_compressed

    @property
    def is_volume(self) -> bool:
        return self.is_valid_image and ((self.NumberOfFrames or 1) > 1)

    @cached_property
    def is_specimen(self) -> bool:
        return "specimen" in (self.StudyDescription or "").lower()

    @property
    def image_area(self) -> Optional[int]:
        if self.Rows is not None and self.Columns is not None:
            return self.Rows * self.Columns
        return None

    @cached_property
    def is_magnified(self) -> bool:
        keywords = {"magnification", "magnified"}
        for modifier in self.view_modifier_codes:
            meaning = get_value(modifier, Tag.CodeMeaning, "").strip().lower()
            if meaning in keywords:
                return True
        return False

    @property
    def pixel_spacing(self) -> Optional[PixelSpacingClass]:
        if not self.PixelSpacing and not self.ImagerPixelSpacing:
            return None
        return PixelSpacingClass.from_tags(
            {
                Tag.PixelSpacing: self.PixelSpacing,
                Tag.ImagerPixelSpacing: self.ImagerPixelSpacing,
            }
        )

    @property
    def view_codes(self) -> Iterator[Dataset]:
        r"""Returns an iterator over all view codes"""
        if self.ViewCodeSequence is not None:
            yield from iterate_view_modifier_codes(self.ViewCodeSequence)

    @property
    def view_code_meanings(self) -> List[str]:
        r"""Returns a list of view code meanings"""
        return [get_value(modifier, Tag.CodeMeaning, "").strip().lower() for modifier in self.view_codes]

    @property
    def view_modifier_codes(self) -> Iterator[Dataset]:
        r"""Returns an iterator over all view modifier codes"""
        if self.ViewModifierCodeSequence is not None:
            yield from iterate_view_modifier_codes(self.ViewModifierCodeSequence)

    @property
    def view_modifier_code_meanings(self) -> List[str]:
        r"""Returns a list of view modifier code meanings"""
        return [get_value(modifier, Tag.CodeMeaning, "").strip().lower() for modifier in self.view_modifier_codes]

    @cached_property
    def all_view_code_meanings(self) -> Set[str]:
        r"""Returns a set of code meanings for all view or view modifier codes"""
        return {meaning for meaning in chain(self.view_code_meanings, self.view_modifier_code_meanings)}

    @classmethod
    def from_dicom(
        cls: Type[R],
        path: PathLike,
        dcm: Dicom,
        helpers: Iterable["RecordHelper"] = [],
        **overrides,
    ) -> R:
        r"""Creates a :class:`MammogramFileRecord` from a DICOM file.

        Args:
            path: Path to DICOM file (needed to set ``path`` attribute)
            dcm: Dicom file object
            modality: Optional modality override
        """
        # raise an exception if Rows/Columns aren't present
        for tag in (Tag.Rows, Tag.Columns):
            value = get_value(dcm, tag, None)
            if value is None:
                raise DicomKeyError(tag)
            elif not value:
                raise DicomValueError(tag)
        rec = cast(R, super().from_dicom(path, dcm, **overrides))
        rec = apply_helpers(rec.path, rec, helpers)
        return rec

    def to_dict(self) -> Dict[str, Any]:
        result = super().to_dict()
        result["magnified"] = self.is_magnified
        if self.ViewCodeSequence is not None:
            result["ViewCodeSequence"] = self.view_code_meanings
        if self.ViewModifierCodeSequence is not None:
            result["ViewCodeModifierSequence"] = self.view_modifier_code_meanings
        return result

    @classmethod
    def _from_dict(cls: Type[R], target: Dict[str, Any]) -> R:
        result = super()._from_dict(target)

        # convert view code meaning strings in dict to Sequence
        view_codes = target.get("ViewCodeSequence", [])
        if view_codes:
            result = replace(result, ViewCodeSequence=DicomImageFileRecord.make_view_code_sequence(view_codes))
        view_modifier_codes = target.get("ViewCodeModifierSequence", [])
        if view_modifier_codes:
            result = replace(
                result, ViewModifierCodeSequence=DicomImageFileRecord.make_view_code_sequence(view_modifier_codes)
            )

        return cast(R, result)

    @staticmethod
    def make_view_code(meaning: str) -> Dataset:
        vc = Dataset()
        vc[Tag.CodeMeaning] = DataElement(Tag.CodeMeaning, "ST", meaning)
        return vc

    @staticmethod
    def make_view_code_sequence(meanings: Iterable[str]) -> Sequence:
        return Sequence([DicomImageFileRecord.make_view_code(meaning) for meaning in meanings])

    def standardized_filename(self, file_id: Optional[str] = None) -> StandardizedFilename:
        path = super().standardized_filename(file_id)
        if self.is_specimen:
            path = path.add_modifier("specimen")
        return path


@RECORD_REGISTRY(name="mammogram", suffixes=[".dcm"])
@dataclass(frozen=True, order=False, eq=False)
class MammogramFileRecord(DicomImageFileRecord):
    r"""Data structure for storing critical information about a DICOM file.
    File IO operations on DICOMs can be expensive, so this class collects all
    required information in a single pass to avoid repeated file opening.
    """

    mammogram_type: MammogramType = MammogramType.UNKNOWN
    view_position: ViewPosition = ViewPosition.UNKNOWN
    laterality: Laterality = Laterality.UNKNOWN
    PaddleDescription: Optional[str] = None
    BreastImplantPresent: Optional[str] = None

    def __post_init__(self):
        super().__post_init__()
        if isinstance(self.mammogram_type, str):
            object.__setattr__(self, "mammogram_type", MammogramType.from_str(self.mammogram_type))
        if isinstance(self.view_position, str):
            object.__setattr__(self, "view_position", ViewPosition.from_str(self.view_position))
        if isinstance(self.laterality, str):
            object.__setattr__(self, "laterality", Laterality.from_str(self.laterality))

    def __lt__(self, other: FileRecord) -> bool:
        return self.is_preferred_to(other)

    def __gt__(self, other: FileRecord) -> bool:
        if isinstance(other, MammogramFileRecord):
            return other.is_preferred_to(self)
        return super().__gt__(other)

    def __le__(self, other: FileRecord) -> bool:
        if isinstance(other, MammogramFileRecord):
            prefer_self = self.is_preferred_to(other)
            prefer_other = other.is_preferred_to(self)
            return prefer_self or not (prefer_self or prefer_other)
        return super().__le__(other)

    def __ge__(self, other: FileRecord) -> bool:
        if isinstance(other, MammogramFileRecord):
            prefer_self = self.is_preferred_to(other)
            prefer_other = other.is_preferred_to(self)
            return prefer_other or not (prefer_self or prefer_other)
        return super().__ge__(other)

    def is_preferred_to(self, other: FileRecord) -> bool:
        r"""Checks if this record is preferred to another record.

        Preference is determined as follows:
            * If the other record is not a mammogram, fall back to comparison by SOPInstanceUID
            * If this record is a standard view and the other record is not, this record is preferred
            * If this record is an implant displaced view and the other record is not, this record is preferred
            * If this record's image type is preferred over the other record's image type, this record is preferred
            * If this record has a higher resolution than the other record, this record is preferred
        """
        if isinstance(other, MammogramFileRecord):
            # Standard views take priority over nonstandard views
            if self.is_standard_mammo_view and not other.is_standard_mammo_view:
                return True
            # Implant displaced views take priority over implant views (only for same study)
            elif (
                self.StudyInstanceUID == other.StudyInstanceUID
                and self.is_implant_displaced
                and not other.is_implant_displaced
            ):
                return True
            # If mammograms have different types, order by type priority
            elif (
                self.mammogram_type is not None
                and other.mammogram_type is not None
                and self.mammogram_type != other.mammogram_type
            ):
                return self.mammogram_type.is_preferred_to(other.mammogram_type)
            # If mammograms have different resolutions, order by resolution
            elif self.image_area != other.image_area:
                # Higher resolution is preferred, so flip the comparison sign
                return (self.image_area or float("inf")) > (other.image_area or float("inf"))
        # Super compares by SOPInstanceUID
        return super().__lt__(other)

    @property
    def mammogram_view(self) -> MammogramView:
        return MammogramView.create(self.laterality, self.view_position)

    @property
    def is_2d(self) -> bool:
        return self.mammogram_type is not None and self.mammogram_type != MammogramType.TOMO

    @cached_property
    def is_spot_compression(self) -> bool:
        if any(s in (self.PaddleDescription or "") for s in ("SPOT", "SPT")):
            return True
        if "spot" in (self.ViewPosition or "").lower():
            return True
        return "spot compression" in self.all_view_code_meanings

    @cached_property
    def is_magnified(self) -> bool:
        return "MAG" in (self.PaddleDescription or "") or super().is_magnified

    @cached_property
    def is_implant_displaced(self) -> bool:
        return "implant displaced" in self.all_view_code_meanings

    @cached_property
    def is_tangential(self) -> bool:
        if "tan" in (self.ViewPosition or "").lower():
            return True
        return "tangential" in self.all_view_code_meanings

    @cached_property
    def is_nipple_in_profile(self) -> bool:
        if "np" in (self.ViewPosition or "").lower():
            return True
        return "nipple in profile" in self.all_view_code_meanings

    @cached_property
    def is_infra_mammary_fold(self) -> bool:
        if "imf" in (self.ViewPosition or "").lower():
            return True
        return "infra-mammary fold" in self.all_view_code_meanings

    @cached_property
    def is_anterior_compression(self) -> bool:
        if "ac" in (self.ViewPosition or "").lower():
            return True
        return "anterior compression" in self.all_view_code_meanings

    @cached_property
    def is_stereo(self) -> bool:
        r"""Check if this is a stereotactic biopsy mammogram"""
        if self.ImageType and "STEREO" in self.ImageType:
            return True

        # NOTE: this info is also in the PerformedProtocolCodeSequence.
        # However, checking the 3 places below should be sufficient
        return any(
            "stereo" in (getattr(self, tag.name) or "").lower()
            for tag in (
                Tag.StudyDescription,
                Tag.SeriesDescription,
                Tag.PerformedProcedureStepDescription,
            )
        )

    @cached_property
    def is_standard_mammo_view(self) -> bool:
        r"""Checks if this record corresponds to a standard mammography view.
        Standard mammography views are the MLO and CC views.
        """
        return (
            (self.view_position in {ViewPosition.MLO, ViewPosition.CC})
            and not self.is_spot_compression
            and not self.is_magnified
            and not self.is_secondary_capture
            and not self.is_for_processing
            and not self.is_cad
            and not self.is_stereo
            and not self.is_anterior_compression
            and not self.is_nipple_in_profile
            and not self.is_infra_mammary_fold
            and not self.is_tangential
            and not self.is_specimen
        )

    @property
    def has_implant(self) -> bool:
        return self.BreastImplantPresent == "YES"

    @classmethod
    def get_standard_mammo_view_lookup(
        cls, records: Iterable["MammogramFileRecord"]
    ) -> Dict[MammogramView, List["MammogramFileRecord"]]:
        needed_views: Dict[MammogramView, List[MammogramFileRecord]] = defaultdict(list)
        for rec in records:
            # only consider standard views
            if not isinstance(rec, MammogramFileRecord) or not rec.is_standard_mammo_view:
                continue
            key = rec.mammogram_view
            if key in STANDARD_MAMMO_VIEWS:
                needed_views[key].append(rec)
        return needed_views

    @classmethod
    def is_complete_mammo_case(cls, records: Iterable["MammogramFileRecord"]) -> bool:
        view_lookup = cls.get_standard_mammo_view_lookup(records)
        return set(view_lookup.keys()) == STANDARD_MAMMO_VIEWS

    @classmethod
    def collection_laterality(cls, records: Iterable["MammogramFileRecord"]) -> Laterality:
        left = right = False
        for rec in records:
            if not isinstance(rec, MammogramFileRecord) or rec.is_secondary_capture:
                continue
            if rec.laterality == Laterality.LEFT:
                left = True
            elif rec.laterality == Laterality.RIGHT:
                right = True

        if left and right:
            return Laterality.BILATERAL
        elif left:
            return Laterality.LEFT
        elif right:
            return Laterality.RIGHT
        else:
            return Laterality.NONE

    @classmethod
    def get_preferred_views(
        cls, col: Iterable["MammogramFileRecord"]
    ) -> Dict[MammogramView, Optional["MammogramFileRecord"]]:
        r"""Selects preferred inference views from a collection of :class:`MammogramFileRecord`s.

        Args:
            col: Collection to select from

        Returns:
            Dictionary giving the selected view (or `None` if a view could not be found) for each of the
            four standard view positions.
        """
        result: Dict[MammogramView, Optional["MammogramFileRecord"]] = {}
        col = list(col)

        # Try each standard view
        for mammo_view in STANDARD_MAMMO_VIEWS:
            lat, view_pos = mammo_view

            # Select only views that match what we're looking for.
            # We permit MLO-like or CC-like views
            def check_is_candidate(rec: MammogramFileRecord) -> bool:
                candidate = rec.mammogram_view
                assert view_pos.is_mlo_like or view_pos.is_cc_like
                laterality_match = candidate.laterality == lat
                view_match = candidate.is_mlo_like if view_pos.is_mlo_like else candidate.is_cc_like
                return view_match and laterality_match

            candidates = list(filter(check_is_candidate, col))

            # Select the most preferred image if one exists
            selection = min(candidates, default=None)
            result[mammo_view] = selection
        return result

    def get_opposing_laterality(self, col: Iterable["MammogramFileRecord"]) -> Optional["MammogramFileRecord"]:
        r"""Selects a complementary view of the opposing laterality from a collection of :class:`MammogramFileRecord`s.
        For example, the complementary view of a left MLO is a right MLO.

        Args:
            col: Collection to select from

        Returns:
            The selected view (or `None` if a view could not be found).
        """
        valid_laterality = self.laterality in {Laterality.LEFT, Laterality.RIGHT}
        valid_view = not (self.view_position is None or self.view_position.is_unknown)
        if not (valid_laterality and valid_view):
            return

        # Target laterality is opposite of this view
        target_laterality = cast(Laterality, self.laterality).opposite
        # Target view is same as this view
        target_view = cast(ViewPosition, self.view_position)

        candidates = self.get_preferred_views(
            {
                rec
                for rec in col
                if isinstance(rec, MammogramFileRecord)
                and rec.laterality == target_laterality
                and rec.view_position == target_view
            }
        )
        result = candidates.get(MammogramView(target_laterality, target_view), None)
        return cast(Optional[MammogramFileRecord], result)

    def standardized_filename(self, file_id: Optional[str] = None) -> StandardizedFilename:
        r"""Returns a standardized filename for the DICOM represented by this :class:`DicomFileRecord`.
        File name will be of the form ``{file_type}_{modifiers}_{view}_{file_id}.dcm``.

        Args:
            file_id:
                A unique identifier for this file that will be added as a postfix to the filename.
                If not provided the output of :func:`get_image_uid()` will be used.
        """
        if self.mammogram_type not in (None, MammogramType.UNKNOWN):
            filetype = self.mammogram_type.simple_name
        else:
            filetype = self.Modality.lower() if self.Modality else "uk"

        # modifiers
        # TODO make this a loop/lookup
        path = super().standardized_filename(file_id)
        modifiers: List[str] = path.prefix_parts[1:]
        if self.is_spot_compression:
            modifiers.append("spot")
        if self.is_magnified:
            modifiers.append("mag")
        if self.is_implant_displaced:
            modifiers.append("id")
        if self.is_tangential:
            modifiers.append("tan")
        if self.is_nipple_in_profile:
            modifiers.append("np")
        if self.is_infra_mammary_fold:
            modifiers.append("imf")
        if self.is_anterior_compression:
            modifiers.append("ac")
        if self.is_stereo:
            modifiers.append("stereo")

        view = f"{self.laterality.short_str}{self.view_position.short_str}"

        prefix = [filetype, view, *modifiers]
        prefix = [p for p in prefix if p]
        path = path.with_prefix(*prefix)
        return path.with_suffix(".dcm")

    @classmethod
    def get_required_tags(cls) -> Set[Tag]:
        return {
            *super().get_required_tags(),
            *Laterality.get_required_tags(),
            *ViewPosition.get_required_tags(),
            *MammogramType.get_required_tags(),
        }

    @classmethod
    def from_dicom(
        cls,
        path: PathLike,
        dcm: Dicom,
        is_sfm: bool = False,
        helpers: Iterable["RecordHelper"] = [],
        **overrides,
    ) -> "MammogramFileRecord":
        r"""Creates a :class:`MammogramFileRecord` from a DICOM file.

        Args:
            path: Path to DICOM file (needed to set ``path`` attribute)
            dcm: Dicom file object
            modality: Optional modality override
            is_sfm: Manual indicator if the mammogram is SFM instead of FFDM
        """
        result = super().from_dicom(path, dcm, **overrides)
        assert isinstance(result, cls)

        modality = result.Modality
        if modality is None:
            raise DicomKeyError(Tag.Modality)
        elif modality != "MG":
            raise DicomValueError(
                f"Modality {modality} is invalid for mammograms. If you are certain {path} is a mammogram, pass "
                "`Modality`='MG' to `from_dicom` or `overrides={'Modality': 'MG'}` to `from_file`."
            )

        laterality = Laterality.from_dicom(dcm)
        view_position = ViewPosition.from_dicom(dcm)
        # ignore modality here because it was checked above
        mammogram_type = MammogramType.from_dicom(dcm, is_sfm, ignore_modality=True)

        result = result.replace(
            laterality=laterality,
            view_position=view_position,
            mammogram_type=mammogram_type,
        )
        result = apply_helpers(result.path, result, helpers)
        return result

    def to_dict(self) -> Dict[str, Any]:
        result = super().to_dict()
        result["mammogram_type"] = self.mammogram_type.simple_name
        result["laterality"] = self.laterality.short_str
        result["view_position"] = self.view_position.short_str
        result["spot_compression"] = self.is_spot_compression
        result["implant_displaced"] = self.is_implant_displaced
        result["standard_mammo_view"] = self.is_standard_mammo_view
        return result


@dataclass
class RecordHelper(ABC):
    r"""A :class:`RecordHelper` implements logic that is run during :class:`FileRecord`
    creation and populates fields using custom logic.
    """

    def __call__(self, path: PathLike, rec: R) -> R:
        r"""Applies postprocessing logic to a :class:`FileRecord`.

        Args:
            path:
                Path of the
        """
        return rec

    def on_read(self, f: T, record_type: Type[FileRecord]) -> T:
        r"""Applies preprocessing logic to the object returned by :func:`FileRecord.read`.

        Args:
            f:
                Object returned by :func:`FileRecord.read`

            record_type:
                Type of :class:`FileRecord` performing the read

        Returns:
            Augmented ``f``
        """
        return f


def apply_helpers(path: PathLike, rec: R, helpers: Iterable[RecordHelper]) -> R:
    for h in helpers:
        if not isinstance(h, RecordHelper):
            raise TypeError(f"type {type(h)} is not a `RecordHelper`")
        logger.debug(f"Applying helper {type(h)} to {path}")
        rec = h(path, cast(R, rec))
        if not isinstance(rec, FileRecord):
            raise TypeError(f"Expected `FileRecord` from {type(h)}, got {type(rec)}")
    return cast(R, rec)


def apply_read_helpers(obj: T, record_type: Type[FileRecord], helpers: Iterable[RecordHelper]) -> T:
    for h in helpers:
        if not isinstance(h, RecordHelper):
            raise TypeError(f"type {type(h)} is not a `RecordHelper`")
        logger.debug(f"Applying read helper {type(obj)}")
        obj = h.on_read(obj, record_type)
    return obj


@HELPER_REGISTRY(name="patient-id-from-path")
@dataclass
class PatientIDFromPath(RecordHelper):
    r"""Helper that extracts a PatientID from the filepath.
    PatientID will be extracted as ``rec.path.parents[helper.level].name``.

    Args:
        Level in the filepath at which to extract PatientID
    """

    def __init__(self, level: int = 0):
        self.level = int(level)

    def __call__(self, path: PathLike, rec: R) -> R:
        if isinstance(rec, SupportsPatientID):
            name = Path(path).parents[self.level].name
            rec = cast(R, rec.replace(PatientID=name))
        return rec


@HELPER_REGISTRY(name="study-date-from-path")
@dataclass
class StudyDateFromPath(RecordHelper):
    r"""Helper that extracts StudyDate from the filepath.
    Study year will be extracted as ``int(rec.path.parents[helper.level].name)``, and
    the StudyDate field will be assigned as ``{year}0101``.

    Args:
        Level in the filepath at which to extract StudyDate
    """

    def __init__(self, level: int = 0):
        self.level = int(level)

    def __call__(self, path: PathLike, rec: R) -> R:
        if isinstance(rec, SupportsStudyDate):
            year = Path(path).parents[self.level].name
            date = f"{year}0101"
            rec = cast(R, rec.replace(StudyDate=date))
        return rec


@HELPER_REGISTRY(name="patient-orientation")
@dataclass
class ParsePatientOrientation(RecordHelper):
    def __call__(self, path: PathLike, rec: R) -> R:
        if isinstance(rec, MammogramFileRecord):
            po_laterality = Laterality.from_patient_orientation(rec.PatientOrientation or [])
            po_view_pos = ViewPosition.from_patient_orientation(rec.PatientOrientation or [])
            rec = cast(
                R,
                rec.replace(
                    laterality=rec.laterality or po_laterality,
                    view_position=rec.view_position or po_view_pos,
                ),
            )
        return rec


# register helpers with some typical values for `level`
for i in range(LEVELS_TO_REGISTER := 3):
    HELPER_REGISTRY(partial(PatientIDFromPath, level=i + 1), name=f"patient-id-from-path-{i + 1}")
    HELPER_REGISTRY(partial(StudyDateFromPath, level=i + 1), name=f"study-date-from-path-{i + 1}")


@dataclass
class DirectoryHelper(RecordHelper):
    def glob(self, path: PathLike, files_only: bool = True) -> Iterator[Path]:
        path = Path(path)
        assert path.is_dir()
        for p in path.glob("*"):
            if not files_only or p.is_file():
                yield p


@HELPER_REGISTRY(name="spot-compression")
@dataclass
class SpotCompressionHelper(DirectoryHelper):
    def __init__(self, size_delta: int = int(1e6)):
        self.size_delta = size_delta

    def __call__(self, path: PathLike, rec: R) -> R:
        if isinstance(rec, MammogramFileRecord):
            path = Path(path)
            file_sizes = {p: size for p, size in self.file_sizes(path.parent).items() if p.suffix == ".dcm"}
            if len(file_sizes) > 4 and (is_spot := rec.file_size < mode(file_sizes.values()) - self.size_delta):
                rec = cast(
                    R,
                    rec.replace(PaddleDescription=(rec.PaddleDescription or "") + " HELPER SPOT"),
                )
        return cast(R, rec)

    def file_sizes(self, path: PathLike) -> Dict[Path, int]:
        return {p: p.stat().st_size for p in self.glob(path)}


@HELPER_REGISTRY(name="modality")
@dataclass
class ModalityHelper(RecordHelper):
    r"""Helper to correct the modality of DICOM object. Some mammograms have a modality other than MG,
    which results in a record other than :class:`MammogramFileRecord` being used.
    """

    force: bool = False

    def on_read(self, f: T, record_type: Type[FileRecord]) -> T:
        if isinstance(f, Dicom) and (self.force or self._maybe_mammogram(f)):
            f.Modality = "MG"
        return cast(T, f)

    def _maybe_mammogram(self, dcm: Dicom) -> bool:
        if get_value(dcm, Tag.Modality, "").strip().lower() == "mg":
            return True
        body_part = get_value(dcm, Tag.BodyPartExamined, "").strip().lower()
        study_desc = get_value(dcm, Tag.StudyDescription, "").strip().lower()
        series_desc = get_value(dcm, Tag.SeriesDescription, "").strip().lower()
        return (body_part == "breast") or ("mammo" in study_desc + series_desc)



#!/usr/bin/env python
# -*- coding: utf-8 -*-


import time

import numpy as np
import pydicom
import pydicom.uid as puid
import pytest
from numpy.random import default_rng
from pydicom.pixels.decoders.base import get_decoder
from pydicom.uid import ImplicitVRLittleEndian, RLELossless

import dicom_utils
from dicom_utils import KeepVolume, ReduceVolume, SliceAtLocation, UniformSample, VolumeHandler, read_dicom_image
from dicom_utils.dicom import data_handlers, default_data_handlers, is_inverted, nvjpeg_decompress, set_pixels


@pytest.fixture
def pynvjpeg():
    return pytest.importorskip("pynvjpeg", reason="pynvjpeg is not installed")


class TestReadDicomImage:
    @pytest.mark.parametrize(
        "tsuid",
        [
            puid.ExplicitVRLittleEndian,
            puid.ImplicitVRLittleEndian,
            puid.JPEG2000Lossless,
            *puid.JPEGLSTransferSyntaxes,
        ],
    )
    def test_decoder(self, tsuid):
        assert get_decoder(tsuid) is not None

    def test_shape(self, dicom_object):
        array = read_dicom_image(dicom_object)
        assert isinstance(array, np.ndarray)
        assert array.ndim == 3, "dims C x H x W"
        assert array.shape[0] == 1, "channel dim size == 1"
        assert array.shape[1] == 128, "height dim size == 128"
        assert array.shape[2] == 128, "width dim size == 128"

    def test_array_dtype(self, dicom_object):
        array = read_dicom_image(dicom_object, rescale=False)
        assert isinstance(array, np.ndarray)
        assert array.dtype == np.int16

    def test_min_max_values(self, dicom_object):
        array = read_dicom_image(dicom_object, rescale=False)
        assert isinstance(array, np.ndarray)
        assert array.min() == 128, "min pixel value 128"
        assert array.max() == 2191, "max pixel value 2191"

    def test_invalid_TransferSyntaxUID_loose_interpretation(self, dicom_object):
        dicom_object.file_meta.TransferSyntaxUID = "1.2.840.10008.1.2.4.90"  # Assign random invalid TransferSyntaxUID
        array = read_dicom_image(dicom_object, use_nvjpeg=False, rescale=False)
        assert isinstance(array, np.ndarray)
        assert array.min() == 128, "min pixel value 128"
        assert array.max() == 2191, "max pixel value 2191"

    def test_invalid_TransferSyntaxUID_exception(self, dicom_object):
        dicom_object.file_meta.TransferSyntaxUID = "1.2.840.10008.1.2.4.90"  # Assign random invalid TransferSyntaxUID
        with pytest.raises(ValueError) as e:
            read_dicom_image(dicom_object, strict_interp=True, use_nvjpeg=False)
        assert "does not appear to be correct" in str(e), "The expected exception message was not returned."

    def test_invalid_PixelData(self, dicom_object):
        dicom_object.PixelData = b""
        with pytest.raises(ValueError) as e:
            read_dicom_image(dicom_object, use_nvjpeg=False)
        expected_msg = "Unable to parse the pixel array after trying all possible TransferSyntaxUIDs."
        assert expected_msg in str(e), "The expected exception message was not returned."

    def test_missing_PixelData(self, dicom_object):
        del dicom_object.PixelData
        with pytest.raises(AttributeError) as e:
            read_dicom_image(dicom_object, use_nvjpeg=False)
        expected_msg = "The dataset has no 'Pixel Data', 'Float Pixel Data' or 'Double Float Pixel Data' element, no pixel data to decode"
        assert expected_msg in str(e), "The expected exception message was not returned."

    @pytest.mark.parametrize("shape_override", [None, (32, 32), (32, 32, 32)])
    def test_stop_before_pixels(self, dicom_object, shape_override):
        np.random.seed(42)
        array1 = read_dicom_image(dicom_object)
        array2 = read_dicom_image(dicom_object, stop_before_pixels=True, override_shape=shape_override)
        assert isinstance(array1, np.ndarray)
        assert isinstance(array2, np.ndarray)

        if shape_override is None:
            assert not (array1 == array2).all()
            assert array1.shape == array2.shape
        else:
            assert array2.shape == (1,) + shape_override

    @pytest.mark.parametrize(
        "handler",
        [
            pytest.param(lambda x: x, marks=pytest.mark.xfail(raises=TypeError)),
            KeepVolume(),
            SliceAtLocation(4),
            UniformSample(4, method="count"),
            ReduceVolume(output_frames=4),
        ],
    )
    def test_volume_handling(self, dicom_object_3d, handler, mocker, transfer_syntax):
        # NOTE: we need to patch handle_dicom for VolumeHandlers, and __call__ for other cases
        # There's an issue with mocking __call__ in all cases, possibly due to the use of a functools.partial
        if isinstance(handler, VolumeHandler):
            spy = mocker.patch.object(handler, "handle_dicom", wraps=handler.handle_dicom)
        else:
            spy = mocker.patch.object(handler, "__call__", wraps=handler)
        F = 8
        dcm = dicom_object_3d(num_frames=F, syntax=transfer_syntax)
        array1 = read_dicom_image(dcm, volume_handler=handler, strict_interp=True)
        spy.assert_called_once()
        assert spy.mock_calls[0].args[0] == dcm, "handler should be called with DICOM object"
        assert array1.ndim < 4 or array1.shape[1] != 1, "3D dim should be squeezed when D=1"

    @pytest.mark.parametrize("has_voi_lut", [True, False])
    def test_convert_3d_voi_lut(self, mocker, dicom_object_3d, has_voi_lut):
        dcm = dicom_object_3d(num_frames=8)
        if has_voi_lut:
            dcm.WindowCenter = 512
            dcm.WindowWidth = 512

        spy = mocker.spy(dicom_utils.dicom, "convert_frame_voi_lut")
        read_dicom_image(dcm, strict_interp=True)
        if has_voi_lut:
            spy.assert_not_called()
        else:
            spy.assert_called_once()

    def test_decoding_speed(self, dicom_file_j2k: str) -> None:
        # Make sure that our set of pixel data handlers is actually faster than the default set

        def time_decode() -> float:
            start_time = time.time()
            pydicom.dcmread(dicom_file_j2k).pixel_array
            return time.time() - start_time

        pydicom.config.pixel_data_handlers = default_data_handlers  # type: ignore
        default_decode_time = 2 * time_decode()

        pydicom.config.pixel_data_handlers = data_handlers  # type: ignore
        decode_time = time_decode()

        assert decode_time < default_decode_time, f"{decode_time} is not less than {default_decode_time}"

    def test_as_uint8(self, dicom_object):
        array = read_dicom_image(dicom_object, as_uint8=True)
        assert isinstance(array, np.ndarray)
        assert array.dtype == np.uint8
        assert array.min() == 0
        assert array.max() == 255

    def test_read_rgb(self, dicom_object):
        test_file = pydicom.data.get_testdata_file("SC_rgb_rle.dcm")  # type: ignore
        dicom_object = pydicom.dcmread(test_file)
        array = read_dicom_image(dicom_object)
        assert isinstance(array, np.ndarray)
        assert array.shape[0] == 3
        assert array.shape[-2:] == (100, 100)
        assert array.dtype == np.uint8

    @pytest.mark.parametrize(
        "dicom_fixture,use_nvjpeg,available,exp",
        [
            # 3D JPEG2000
            pytest.param("dicom_file_j2k_3d_uint16", True, True, True, marks=pytest.mark.usefixtures("pynvjpeg")),
            pytest.param("dicom_file_j2k_3d_uint16", None, True, True, marks=pytest.mark.usefixtures("pynvjpeg")),
            pytest.param("dicom_file_j2k_3d_uint16", True, False, True, id="cpu-fallback"),
            pytest.param("dicom_file_j2k_3d_uint16", None, False, False),
            pytest.param("dicom_file_j2k_3d_uint16", False, False, False),
            # 2D JPEG2000
            pytest.param("dicom_file_j2k_uint16", True, True, True, marks=pytest.mark.usefixtures("pynvjpeg")),
            pytest.param("dicom_file_j2k_uint16", None, True, True, marks=pytest.mark.usefixtures("pynvjpeg")),
            pytest.param("dicom_file_j2k_uint16", True, False, True, id="cpu-fallback"),
            pytest.param("dicom_file_j2k_uint16", None, False, False),
            pytest.param("dicom_file_j2k_uint16", False, False, False),
            # not JPEG2000, we should not attempt nvJPEG2K
            pytest.param("dicom_file_jpl14", True, True, False),
            pytest.param("dicom_file_jpl14", None, True, False),
            pytest.param("dicom_file_jpl14", True, False, False),
            pytest.param("dicom_file_jpl14", None, False, False),
            pytest.param("dicom_file_jpl14", False, False, False),
            # We should never run on int16 because it's not supported
            pytest.param("dicom_file_j2k_int16", True, True, False),
            pytest.param("dicom_file_j2k_int16", None, True, False),
            pytest.param("dicom_file_j2k_int16", True, False, False),
            pytest.param("dicom_file_j2k_int16", None, False, False),
            pytest.param("dicom_file_j2k_int16", False, False, False),
        ],
    )
    def test_nvjpeg_autoselect(self, request, mocker, dicom_fixture, use_nvjpeg, available, exp):
        dicom_file = request.getfixturevalue(dicom_fixture)

        # patch methods
        a = mocker.patch("dicom_utils.dicom.nvjpeg2k_is_available", return_value=available)

        # patch to use CPU instead of GPU
        def new(dcm, *args, **kwargs):
            if not a():
                raise ImportError("nvJPEG not available")
            return dcm.pixel_array

        m = mocker.patch("dicom_utils.dicom.nvjpeg_decompress", side_effect=new)

        # read and check
        with pydicom.dcmread(dicom_file) as dcm:
            read_dicom_image(dcm, use_nvjpeg=use_nvjpeg, nvjpeg_batch_size=1)
        assert m.called == exp

    @pytest.mark.ci_skip  # CircleCI will not have a GPU
    @pytest.mark.usefixtures("pynvjpeg")
    @pytest.mark.parametrize(
        "dicom_fixture",
        [
            "dicom_file_j2k_3d_uint16",
            "dicom_file_j2k_uint16",
            "dicom_file_jpl14",
            "dicom_file_j2k_int16",
        ],
    )
    def test_nvjpeg(self, request, dicom_fixture):
        dicom_file_j2k = request.getfixturevalue(dicom_fixture)

        ds = pydicom.dcmread(dicom_file_j2k)
        gpu_image = read_dicom_image(ds, use_nvjpeg=True, nvjpeg_batch_size=1)
        cpu_image = read_dicom_image(ds, use_nvjpeg=False)

        assert cpu_image.shape == gpu_image.shape
        assert (cpu_image == gpu_image).all()

    @pytest.mark.parametrize(
        "presentation,exp",
        [
            ("FOR PROCESSING", True),
            ("FOR PRESENTATION", True),
        ],
    )
    def test_for_algorithm_presentation_handling(self, mocker, dicom_object, presentation, exp):
        m1 = mocker.patch("dicom_utils.dicom.invert_color", return_value=dicom_object.pixel_array)
        dicom_object.PresentationIntentType = presentation
        dicom_object.PhotometricInterpretation = "MONOCHROME1"
        read_dicom_image(dicom_object)
        assert m1.called == exp, "invert_color not called as expected"

    def test_voi_lut_control(self, mocker, dicom_object):
        np.random.seed(42)
        spy = mocker.spy(dicom_utils.dicom, "apply_voi_lut")
        dicom_object.WindowCenter = 512
        dicom_object.WindowWidth = 512
        array1 = read_dicom_image(dicom_object, voi_lut=True)
        array2 = read_dicom_image(dicom_object, voi_lut=False)
        assert (array1 != array2).any()
        assert spy.call_count == 1

    def test_inversion_control(self, mocker, dicom_object):
        np.random.seed(42)
        spy = mocker.spy(dicom_utils.dicom, "invert_color")
        dicom_object.PhotometricInterpretation = "MONOCHROME1"
        array1 = read_dicom_image(dicom_object, inversion=True)
        array2 = read_dicom_image(dicom_object, inversion=False)
        assert (array1 != array2).any()
        assert spy.call_count == 1

    def test_rescale_control(self, mocker, dicom_object):
        np.random.seed(42)
        spy = mocker.spy(dicom_utils.dicom, "apply_rescale")
        dicom_object.RescaleIntercept = 1
        dicom_object.RescaleSlope = 2
        array1 = read_dicom_image(dicom_object, rescale=True)
        array2 = read_dicom_image(dicom_object, rescale=False)
        assert (array1 != array2).any()
        assert spy.call_count == 1


@pytest.mark.ci_skip  # CircleCI will not have a GPU
@pytest.mark.usefixtures("pynvjpeg")
@pytest.mark.parametrize(
    "dicom_fixture",
    [
        "dicom_file_j2k_3d_uint16",
        "dicom_file_j2k_uint16",
        pytest.param("dicom_file_jpl14", marks=pytest.mark.xfail(raises=ValueError)),
        pytest.param("dicom_file_j2k_int16", marks=pytest.mark.xfail(raises=ValueError)),
    ],
)
def test_nvjpeg_decompress(request, dicom_fixture):
    dicom_file_j2k = request.getfixturevalue(dicom_fixture)

    ds = pydicom.dcmread(dicom_file_j2k)
    gpu_image = nvjpeg_decompress(ds, batch_size=1)
    ds = pydicom.dcmread(dicom_file_j2k)
    cpu_image = ds.pixel_array

    assert cpu_image.shape == gpu_image.shape
    assert (cpu_image == gpu_image).all()


def test_deprecated_is_inverted(dicom_object):
    with pytest.warns(DeprecationWarning):
        assert not is_inverted(dicom_object.PhotometricInterpretation)


@pytest.mark.parametrize(
    "rows,cols,num_frames,bits,orig_tsuid",
    [
        pytest.param(32, 32, 1, 16, ImplicitVRLittleEndian),
        pytest.param(32, 64, 1, 16, ImplicitVRLittleEndian),
        pytest.param(32, 64, 3, 16, ImplicitVRLittleEndian),
        pytest.param(32, 64, 3, 16, RLELossless),
    ],
)
def test_set_pixels(dicom_object, rows, cols, num_frames, bits, transfer_syntax, orig_tsuid):
    dicom_object.Rows = rows
    dicom_object.Columns = cols
    dicom_object.NumberOfFrames = num_frames
    dicom_object.file_meta.TransferSyntaxUID = orig_tsuid

    low = 0
    high = bits
    channels = 1 if dicom_object.PhotometricInterpretation.startswith("MONOCHROME") else 3
    size = tuple(x for x in (channels, num_frames, rows, cols) if x > 1)
    rng = default_rng(seed=42)
    arr = rng.integers(low, high, size, dtype=np.uint16)

    output = set_pixels(dicom_object, arr, transfer_syntax)
    arr_out = output.pixel_array
    assert isinstance(arr_out, np.ndarray)
    assert output.file_meta.TransferSyntaxUID == transfer_syntax
